/**********************************************************************
 * This file is part of FMR (Fast Machine Reassignment)               
 *                                                                   
 * ROADEF/EURO Challenge 2012: Machine Reassignment                 
 * http://challenge.roadef.org/2012/en/				   
 * Team S26
 *     
 * Copyright (C) 2011-12 Laboratoire d'Informatique de Paris Nord 
 * UMR CNRS 7030 
 * Institut Galilée - Université Paris-Nord
 * 99, avenue Jean-Baptiste Clément 
 * 93430 Villetaneuse  
 *        
 * FMR is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *  
 * FMR is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with FMR; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 **********************************************************************/

/* Gestion d'ensemble d'entiers non nuls : 1 bit à 1 signifie élément présent,
 * Ensemble_t est un tableau d'entiers longs */
#include "fmr_ens.h"

/* Alloue la place nécessaire, si maxelt=0 allouera tout de même 1 long,*/
/*---------------------------------------------------------------------------*/
Ensemble_t Ens_nouveau(unsigned int maxelt) {
	return (Ensemble_t)malloc(sizeof(unsigned long)* NB_LONG(maxelt));
}

/*---------------------------------------------------------------------------*/
void Ens_copy(Ensemble_t seta,const Ensemble_t setb, unsigned int maxelt) {
  unsigned int i;

  for (i=0;i<NB_LONG(maxelt);i++)
    seta[i]=setb[i];
}


/*---------------------------------------------------------------------------*/
bool Ens_appartient(Element_t elt,Ensemble_t set) {
  unsigned high,low;

  high= (unsigned) (elt/NB_BITS); /* "partie haute" : l'indice du tab. set */
  low= elt % NB_BITS;             /* "partie basse" : le bit dans le 
                                    mot-memoire de set[high] */
  if (set[high] & (1uL<<low))
    return(true);
  else
    return(false);
}

/*---------------------------------------------------------------------------*/
void Ens_ajoute(Element_t elt,Ensemble_t set) {
  unsigned high,low;

  high= (unsigned) (elt/NB_BITS);
  low= elt % NB_BITS;

  set[high] |= 1uL<<low;
}

/*---------------------------------------------------------------------------*/
void Ens_ote(Element_t elt,Ensemble_t set) {
  unsigned high,low;

  high= (unsigned) (elt/NB_BITS);
  low= elt % NB_BITS;

  set[high] &= ~(1uL<<low);
}

/*---------------------------------------------------------------------------*/
bool Ens_est_vide(Ensemble_t t, unsigned int maxelt) {     
  unsigned int i;

  for(i=0;i<NB_LONG(maxelt) && t[i]==0uL;i++);
  return(i==NB_LONG(maxelt));
}


/*---------------------------------------------------------------------------*/
bool Ens_est_singleton(Ensemble_t t, unsigned int maxelt) {
  unsigned int i;

  for (i=0; i<NB_LONG(maxelt) && t[i]==0L; i++);

  if (i==NB_LONG(maxelt))
	return false;
  if ((t[i] & (t[i]-1)) !=0) /* plus d'un bit à 1 dans cette case */
	return false;

  /* verifie que le reste est a 0 */
  for (; i< NB_LONG(maxelt) && t[i]==0L; i++);

  return (i == NB_LONG(maxelt));
}

/*---------------------------------------------------------------------------*/
void Ens_zero(Ensemble_t set, unsigned int maxelt){
  unsigned int i;

  for(i=0;i<NB_LONG(maxelt);i++)
    set[i]=0uL;
}

/*---------------------------------------------------------------------------*/
/* difficile d'utiliser stdarg ici sinon imposer l'usage des va_list aux
 * utilisateurs ...
 * param1: maxelt.*/
/*VARARGS2*/
unsigned Ens_appli(void (*fonc)(), Ensemble_t set,unsigned int maxelt/*,int param1,int param2,int param3,int param4,int param5,int param6,int param7,int param8,int param9*/)
/*unsigned Ens_appli(void (*fonc)(),Ensemble_t set,...)*/
{
  unsigned i,k,card=0;
  unsigned long elt;

  for (i=0;i<NB_LONG(maxelt); i++)
    for (k=0, elt=set[i] ; k<NB_BITS && elt ; elt >>=1, k++)
        if (elt & 0x01)
        {
            (*fonc)(i*NB_BITS+k/*,param1,param2,param3,param4,param5,param6,param7,param8,param9*/);
            card++;
        }
  return(card);
}

/*---------------------------------------------------------------------------*/
void Ens_print_elt(Element_t);
void Ens_print_elt(Element_t iter) {
    printf("%4d,",iter);
}

unsigned Ens_aff(Ensemble_t set,unsigned int maxelt)
{
  return(Ens_appli(Ens_print_elt,set,maxelt));
}

/*---------------------------------------------------------------------------*/
unsigned Ens_card(Ensemble_t set, unsigned int maxelt)
{
  unsigned long elt;
  unsigned i,total=0;

  for (i=0;i<NB_LONG(maxelt);i++)
      for(elt=set[i];elt;elt &= (elt-1)) /*En notation compl. a 2, j&(j-1) 
                                           detruit le bit le plus a doite de j*/
            total++;
  return(total);
}

/*---------------------------------------------------------------------------*/
Element_t Ens_elt_suivant(Element_t el,Ensemble_t s, unsigned int maxelt)
{
    unsigned high,low;
    unsigned long e,j;
    unsigned int i,k=0;

    high= (unsigned) (el/NB_BITS); /* "partie haute" : l'indice du tab.
                                           set */
    low= el % NB_BITS;             /* "partie basse" : le bit dans le
                                           mot-memoire de set[high] */

    low++;
    if (low< NB_BITS && ((e=s[high]) & ((unsigned long)~0 << low)))
    {
        for(j= 1uL<<low,k=low;k<NB_BITS && !(e & j) ;j<<=1,k++);
        if (k<NB_BITS)
            return(high*NB_BITS+k);
    }
    for (i= high +1;i<NB_LONG(maxelt);i++)
        if ( (e=s[i]) )
        {
            for(j=1,k=0;k<NB_BITS && !(e & j) ;j<<=1,k++);
            if (k<NB_BITS)
                return(i*NB_BITS+k);
        }
    if (i<NB_LONG(maxelt))
        return(i*NB_BITS+k);
    else
        return (0);
}

/*---------------------------------------------------------------------------*/
/* les elts de 0 a maxelt-1 compris sont positionnes, suppose maxelt>=1  */
void Ens_rempli(Ensemble_t set, unsigned int maxelt){
  unsigned int i;

  for(i=0;i<NB_LONG(maxelt) ;i++)
    set[i]= (unsigned long)~0;
  set[NB_LONG(maxelt)-1] = (unsigned long)~0 >> (NB_BITS - (maxelt % NB_BITS));
}


/*---------------------------------------------------------------------------*/
unsigned int Ens_ote_prem_elt(Ensemble_t set, unsigned int maxelt) {
  unsigned int e;

  if (Ens_appartient(0,set))
    return(0);
  if ( (e=Ens_elt_suivant(0,set,maxelt)) ==0)
    return((unsigned int) -1); /* signifiant ens. set est vide */
  else {
    Ens_ote(e,set);
    return(e);
  }
}

