/**********************************************************************
 * This file is part of FMR (Fast Machine Reassignment)               
 *                                                                   
 * ROADEF/EURO Challenge 2012: Machine Reassignment                 
 * http://challenge.roadef.org/2012/en/				   
 * Team S26
 *     
 * Copyright (C) 2011-12 Laboratoire d'Informatique de Paris Nord 
 * UMR CNRS 7030 
 * Institut Galilée - Université Paris-Nord
 * 99, avenue Jean-Baptiste Clément 
 * 93430 Villetaneuse  
 *        
 * FMR is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *  
 * FMR is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *  
 * You should have received a copy of the GNU General Public License
 * along with FMR; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 **********************************************************************/

#include "fmr_exploEjection.h"

#include "fmr.h"
#include "fmr_mem.h"
#include "fmr_tools.h"
#include "fmr_ens.h"
#include "fmr_constraints.h"
#include "fmr_cost.h"
#include "fmr_psmvt.h"
#include "fmr_threads.h"

#include <pthread.h>
#include <string.h>


#define VOISINAGEMAX 100

/* verification minimales pour best_shift: il faut que m_ideale reste un point de chute valide */
bool
verif_dependances (placement_t *placmnt, unsigned short p0, unsigned short m_ideale, unsigned short p, unsigned short m) {
    unsigned short s0= instance.procs[p0].service;
    unsigned short s = instance.procs[p].service;
    unsigned short n_ideale=  instance.machines[m_ideale].neighborhood;
    
    if (n_ideale != instance.machines[m].neighborhood)
	return(true);

    if (s ==s0)
	return (true);
    
   /* voisinage: dependancy constraint */
    if (instance.services[s].nb_succ!=0) {
	if (listeCompacte2D_est_singleton( & placmnt->liste_de_processus_d_un_service_dans_un_voisinage, n_ideale, s)) {
		//DEBUG_PRINT("risque orphelinat\n");
            return (false);
	}
    }
    return(true);
}


/* Recherche du meilleur switch d'une machine m diff. de m0 et de m_ideale
 * telle que le gain soit max.
 * Il faudra peut etre restreindre l'ens. des ps au vu des dépéndances de service
 * susceptibles de bouger vers m0, en attendant for !*/
long
best_shift (const placement_t const *initial, placement_t * placmt,
	    unsigned short m_dest, 
	    //Ensemble_t machines_interdites, 
	    bool *machines_interdites,
	    long gainMin,
	    unsigned short *ps, unsigned short *mach, unsigned short p0, unsigned short m_ideale)
{
    unsigned short m;
    long gain = gainMin, gainTempo;
    liste_t *ptr;
    unsigned short p;

    *ps = NOPROCESS;			/* pas trouvé */
    *mach = NOMACHINE;

    for (m = 0; m < instance.nb_machines; m++) {
	if (! machines_interdites[m]) {
	    p = listeCompacte_tete(&placmt->liste_procs,m);
	    while (p != LISTE_VIDE_US) {
		if (check_all_constraints_shift (placmt, p, m, m_dest) 
		    && verif_dependances (placmt, p0, m_ideale, p, m) 
		    && (gainTempo = calcule_gain_shift (initial, placmt, p, m, m_dest)) > gain) {
		    *ps = p;
		    *mach = m;
		    gain = gainTempo;
		}
		p= listeCompacte_next(&placmt->liste_procs,m,p);
	    }
	}
    }
    return (gain);
}

/* Recherche du meilleur switch d'une machine m diff. de m0 et de m_ideale
 * telle que le gain soit max.
 * Il faudra peut etre restreindre l'ens. des ps au vu les dépéndances de service
 * susceptibles de bouger vers m0, en attendant for !*/
long
best_shift_restreint (const placement_t const *initial, placement_t * placmt,
	    unsigned short m_dest, Ensemble_t machines_interdites, long gainMin,
	    unsigned short *ps, unsigned short *mach, unsigned short p0, unsigned short m_ideale, unsigned short present_ou_indice[])
{
    unsigned short m, indice;
    long gain = gainMin, gainTempo;
    liste_t *ptr;
    unsigned short p;

    *ps = NOPROCESS;			/* pas trouvé */
    *mach = NOMACHINE;

    for (m = 0; m < instance.nb_machines; m++) {
	if (!Ens_appartient (m, machines_interdites)) {
	    p = listeCompacte_tete(&placmt->liste_procs,m);
	    while (p != LISTE_VIDE_US) {
		if ((indice=present_ou_indice[p])!=NOPROCESS) { // ps oté par le glouton partiel, on saute à l'indice suivant
		  if (indice ==0) { // il n'y a plus de ps non oté
			  break;
		  } else 
			  p=indice;
		} // sinon il est bien présent.
		if (check_all_constraints_shift (placmt, p, m, m_dest) 
		    && verif_dependances (placmt, p0, m_ideale, p, m) 
		    && (gainTempo = calcule_gain_shift (initial, placmt, p, m, m_dest)) > gain) {
		    *ps = p;
		    *mach = m;
		    gain = gainTempo;
		}
		p= listeCompacte_next(&placmt->liste_procs,m,p);
	    }
	}
    }
    return (gain);
}

/**
 * Algo qui explore un voisinage par chaine d'ejection. La longueur de la chaîne est
 * d'au plus lgmax_chain_ejection
 * Chaine d'ejection : le ps p0 est oté de m0, on cherche m1 telle qu'un ps de m1 puisse venir
 * sur m0, puis m2 telle qu'un ps puisse aller sur m1.
 * tous les états intermédiaires doivent etre réalisables.
 * les mi sont tous différents. 
 * On réinsère p0 en fin de processus : soit en m_ideale (calculée avant) soit en fin de chaine.
 */
long
explo_voisinageEjectionChain (const placement_t const *initial, placement_t * resultat,
                              unsigned short lgmax_chain_ejection, long gainMin)
{
    placement_t pl_temp;
    unsigned short p0, m0, m, m_ideale, p, m_prev;
    long gain, gainTempo, gainPrec, gainTotal = 0;
    unsigned short profVois;
    bool pas_bloque;
    bool *machines_interdites=(bool *)malloc(sizeof(bool)*instance.nb_machines);
    unsigned short nb_machines_interdites;


    pl_temp = alloc_assignement (instance.nb_procs);
    for (p0 = 0; p0 < instance.nb_procs; p0++) {
	m0 = resultat->assignment[p0];

	if (check_remove_proc (p0, m0, resultat)) {	/* si une fois p0 supprimé la sol. est toujours valide */
	    /* pour tous les shifts possibles de  p0, garder le meilleur => m_ideale. */
	    gain = gainMin;
	    m_ideale = m0;
	    for (m = 0; m < instance.nb_machines; m++) {
		if (m != m0 && check_all_constraints_shift(resultat, p0,m0,m)
		    && (gainTempo = calcule_gain_shift (initial, resultat, p0, m0, m)) > gain) {
		    gain = gainTempo;
		    m_ideale = m;
		}
	    }

	    if (m_ideale != m0) {	/* m_ideale permettra de réinsérér p0 en fin d'ejection faute de mieux */
		/*DEBUG_PRINT("p0=%d, s=%d, pl_temp.service_par_machine[m0=%d]=%d, pl_temp.service_par_machine[m_ideale=%d]=%d\n",
			p0,instance.procs[p0].service,
			m0,pl_temp.service_par_machine[m0][instance.procs[p0].service],
			m_ideale, pl_temp.service_par_machine[m_ideale][instance.procs[p0].service]);*/
	        copier_assignement (resultat, &pl_temp);
		gain = calcule_gain_suppr (initial, &pl_temp, p0, m0);
		oter_proc (initial, p0, m0, &pl_temp);
		//DEBUG_PRINT("Ote proc %d de %d, m_ideale=%d\n",p0,m0,m_ideale);
		m_prev = m0;
		profVois = 0;
		pas_bloque = true;
		memset(machines_interdites, false, instance.nb_machines);
		machines_interdites[m_ideale]=true;
		machines_interdites[m0]=true;
		nb_machines_interdites = 2;
		gainPrec=gain;
		pthread_testcancel();
		/* Il faut boucler ici sur la longueur de chaine d'ejection */
		while (profVois < lgmax_chain_ejection && pas_bloque) {
		    /* chercher un shift valide vers m_prev tel que le gain soit max. */
		    gainTempo = best_shift (initial, &pl_temp, m_prev,
				    machines_interdites, gainMin, &p, &m, p0, m_ideale);
		    if (p == NOPROCESS) {
			pas_bloque = false;
		    } else {
			gain += gainTempo;
			gainPrec= gainTempo;
			DEBUG_PRINT("deplace %d de %d à %d, gain=%ld\n",p, m,m_prev,gain);
#ifdef DEBUGIO
			DEBUG_IO_PRINT(", machines interdites=");
			//Ens_aff(machines_interdites, instance.nb_machines);
			DEBUG_IO_PRINT("\n");
#endif
			deplace_proc (initial, &pl_temp, p, m_prev);
			m_prev = m;
			//Ens_ajoute (m, machines_interdites);
			machines_interdites[m]=true;
			profVois++;
		    }
		}
		if (m_prev == m0) {
		    //DEBUG_PRINT ("chaine ejection de longueur nulle %d(%d):-(\n", p0, m0);
		} else {
		    DEBUG_PRINT ("chaine ejection de longueur %d: p0=%d,m0=%d, m_ideale=%d,",
			 profVois, p0, m0, m_ideale);
		    /* il faut réinsérer p0 soit sur la dernière de la chaine soit sur m_ideale */
		    if (!check_add_proc (initial, p0, m_prev, &pl_temp)) {
			if (!check_add_proc (initial, p0, m_ideale, &pl_temp)) {
			    DEBUG_PRINT ("Pb chaine ejection : m_ideale impossible !p=%d, m0=%d, m_ideale=%d\n",
				 p0, m0, m_ideale);
			    continue;
			} else { // p0 ne peut être placé en fin de chaine d'Ejection, on essaie avec un dep. de moins !

			    DEBUG_PRINT ("gain=%ld+%ld (finalement sur m_ideale, derniergain=%ld)\n",
				 gain, calcule_gain_ajout (initial, &pl_temp, p0, m_ideale), gainPrec);
			    gain += calcule_gain_ajout (initial, &pl_temp, p0, m_ideale);
			    gainTotal += gain;
			    ajouter_proc (initial, p0, m_ideale, &pl_temp);
			    copier_assignement (&pl_temp, resultat);
			    resultat->cout -= gain;
			    if (tempsRestant2()<DERNIERE_MIN)
	    		    	atomic_set_meilleur_placement(resultat,'c');
			}
		    } else if (calcule_gain_ajout (initial, &pl_temp, p0, m_ideale) 
			    <= calcule_gain_ajout (initial, &pl_temp, p0, m_prev)) {
			DEBUG_PRINT ("gain=%ld+%ld (finalement sur %d)\n",
				     gain, calcule_gain_ajout (initial, &pl_temp, p0, m_prev),
				     m_prev);
			gain += calcule_gain_ajout (initial, &pl_temp, p0, m_prev);
			gainTotal += gain;
			ajouter_proc (initial, p0, m_prev, &pl_temp);
			copier_assignement (&pl_temp, resultat);
			resultat->cout -= gain;
			if (tempsRestant2()<DERNIERE_MIN)
	    			atomic_set_meilleur_placement(resultat,'c');
		    } else {
			DEBUG_PRINT ("gain=%ld+%ld (finalement sur m_ideale car meilleure)\n",
			     gain, calcule_gain_ajout (initial, &pl_temp, p0, m_ideale));
			gain += calcule_gain_ajout (initial, &pl_temp, p0, m_ideale);
			gainTotal += gain;
			ajouter_proc (initial, p0, m_ideale, &pl_temp);
			copier_assignement (&pl_temp, resultat);
			resultat->cout -= gain;
			if (tempsRestant2()<DERNIERE_MIN)
	    		  atomic_set_meilleur_placement(resultat,'c');
		    }
		}
	    }			// if (m_ideale!=m0)
	}			// if (check_remove_proc(p0, m0, resultat)) 
    }				// for (p0=0; p0< instance.nb_procs; p0++)
    //resultat->cout -= gainTotal ;
    DEBUG_PRINT ("Gain total Ejection:%ld\n", gainTotal);
    if (machines_interdites)
	free(machines_interdites);
    free_placement(&pl_temp);
    return gainTotal;
}

long
explo_voisinageEjectionChain_restreint (const placement_t const *initial, placement_t * resultat,
                              unsigned short lgmax_chain_ejection, unsigned short present_ou_indice[], long gainMin)
{
    placement_t pl_temp;
    unsigned short p0, m0, m, m_ideale, p, m_prev, indice;
    long gain, gainTempo, gainPrec, gainTotal = 0;
    unsigned short profVois;
    bool pas_bloque;
    Ensemble_t machines_interdites=Ens_nouveau(instance.nb_machines);
    unsigned short nb_machines_interdites;


    pl_temp = alloc_assignement (instance.nb_procs);
    for (p0 = 0; p0 < instance.nb_procs; p0++) {
	if ((indice=present_ou_indice[p0])!=NOPROCESS) { // ps oté par le glouton partiel, on saute à l'indice suivant
		if (indice ==0) { // il n'y a plus de ps non oté
			break;
		} else 
			p0=indice;
	} // sinon il est bien présent.
	m0 = resultat->assignment[p0];
// 	printf("p0=%d et m0=%d\n", p0, m0);

	if (check_remove_proc (p0, m0, resultat)) {	/* si une fois p0 supprimé la sol. est toujours valide */
	    /* pour tous les shifts possibles de  p0, garder le meilleur => m_ideale. */
	    gain = gainMin;
	    m_ideale = m0;
	    for (m = 0; m < instance.nb_machines; m++) {
		if (m != m0 && check_all_constraints_shift(resultat, p0,m0,m)
		    && (gainTempo = calcule_gain_shift (initial, resultat, p0, m0, m)) > gain) {
		    gain = gainTempo;
		    m_ideale = m;
		}
	    }

	    if (m_ideale != m0) {	/* m_ideale permettra de réinsérér p0 en fin d'ejection faute de mieux */
		/*DEBUG_PRINT("p0=%d, s=%d, pl_temp.service_par_machine[m0=%d]=%d, pl_temp.service_par_machine[m_ideale=%d]=%d\n",
			p0,instance.procs[p0].service,
			m0,pl_temp.service_par_machine[m0][instance.procs[p0].service],
			m_ideale, pl_temp.service_par_machine[m_ideale][instance.procs[p0].service]);*/
	        copier_assignement (resultat, &pl_temp);
		gain = calcule_gain_suppr (initial, &pl_temp, p0, m0);
		oter_proc (initial, p0, m0, &pl_temp);
		//DEBUG_PRINT("Ote proc %d de %d, m_ideale=%d\n",p0,m0,m_ideale);
		m_prev = m0;
		profVois = 0;
		pas_bloque = true;
		Ens_zero(machines_interdites, instance.nb_machines);
		Ens_ajoute (m_ideale, machines_interdites);
		Ens_ajoute (m0, machines_interdites);
		nb_machines_interdites = 2;
		gainPrec=gain;
		pthread_testcancel();
		/* Il faut boucler ici sur la longueur de chaine d'ejection */
		while (profVois < lgmax_chain_ejection && pas_bloque) {
		    /* chercher un shift valide vers m_prev tel que le gain soit max. */
		    gainTempo = best_shift_restreint (initial, &pl_temp, m_prev,
				    machines_interdites, gainMin, &p, &m, p0, m_ideale, present_ou_indice);
		    if (p == NOPROCESS) {
			pas_bloque = false;
		    } else {
			gain += gainTempo;
			gainPrec= gainTempo;
			DEBUG_PRINT("deplace %d de %d à %d, gain=%ld\n",p, m,m_prev,gain);
#ifdef DEBUGIO
			DEBUG_IO_PRINT(", machines interdites=");
			Ens_aff(machines_interdites, instance.nb_machines);
			DEBUG_IO_PRINT("\n");
#endif
			deplace_proc (initial, &pl_temp, p, m_prev);
			m_prev = m;
			Ens_ajoute (m, machines_interdites);
			profVois++;
		    }
		}
		if (m_prev == m0) {
		    //DEBUG_PRINT ("chaine ejection de longueur nulle %d(%d):-(\n", p0, m0);
		} else {
		    DEBUG_PRINT ("chaine ejection de longueur %d: p0=%d,m0=%d, m_ideale=%d,",
			 profVois, p0, m0, m_ideale);
		    /* il faut réinsérer p0 soit sur la dernière de la chaine soit sur m_ideale */
		    if (!check_add_proc (initial, p0, m_prev, &pl_temp)) {
			if (!check_add_proc (initial, p0, m_ideale, &pl_temp)) {
			    DEBUG_PRINT ("Pb chaine ejection : m_ideale impossible !p=%d, m0=%d, m_ideale=%d\n",
				 p0, m0, m_ideale);
			    continue;
			} else { // p0 ne peut être placé en fin de chaine d'Ejection, on essaie avec un dep. de moins !

			    DEBUG_PRINT ("gain=%ld+%ld (finalement sur m_ideale, derniergain=%ld)\n",
				 gain, calcule_gain_ajout (initial, &pl_temp, p0, m_ideale), gainPrec);
			    gainTotal +=
				gain + calcule_gain_ajout (initial, &pl_temp, p0, m_ideale);
			    ajouter_proc (initial, p0, m_ideale, &pl_temp);
			    copier_assignement (&pl_temp, resultat);
			}
		    } else if (calcule_gain_ajout (initial, &pl_temp, p0, m_ideale) 
			    <= calcule_gain_ajout (initial, &pl_temp, p0, m_prev)) {
			DEBUG_PRINT ("gain=%ld+%ld (finalement sur %d)\n",
				     gain, calcule_gain_ajout (initial, &pl_temp, p0, m_prev),
				     m_prev);
			gainTotal +=
			    gain + calcule_gain_ajout (initial, &pl_temp, p0, m_prev);
			ajouter_proc (initial, p0, m_prev, &pl_temp);
			copier_assignement (&pl_temp, resultat);
		    } else {
			DEBUG_PRINT ("gain=%ld+%ld (finalement sur m_ideale car meilleure)\n",
			     gain, calcule_gain_ajout (initial, &pl_temp, p0, m_ideale));
			gainTotal +=
			    gain + calcule_gain_ajout (initial, &pl_temp, p0, m_ideale);
			ajouter_proc (initial, p0, m_ideale, &pl_temp);
			copier_assignement (&pl_temp, resultat);
		    }
		}
	    }			// if (m_ideale!=m0)
	}			// if (check_remove_proc(p0, m0, resultat)) 
    }				// for (p0=0; p0< instance.nb_procs; p0++)
    resultat->cout -= gainTotal ;
    DEBUG_PRINT ("Gain total Ejection:%ld\n", gainTotal);
    if (machines_interdites)
	free(machines_interdites);
    free_placement(&pl_temp);
    return gainTotal;
}
